// BetterEditView.h : header file
//
// Copyright (c) 1998 Matt Clay, all rights reserved.
//

#if !defined(BETTEREDITVIEW_H)
#define BETTEREDITVIEW_H

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include <afxext.h>
#include "TextActionUtil.h"

class CBEAction;
class CFindReplaceDialog;

/////////////////////////////////////////////////////////////////////////////
// CBetterEditView view

class CBetterEditView : public CEditView
{
public:
	CBetterEditView();           // protected constructor used by dynamic creation
	DECLARE_DYNCREATE(CBetterEditView)

// Attributes
public:
					// because typing is not atomic, we maintain a
					// last action for this view; most of the time,
					// this will be a typing action, and one which
					// we can continue "typing into"
	CBEAction *		lastAction;

					// last character typed in this edit view
	UINT			lastChar;

protected:
					// a pointer to the action target; usually a
					// ptr. to the document object of this view
	void *			actionTarget;

public:
					// flags for controlling redraws
	bool			redraw;
	BOOL			locked;

	static CFindReplaceDialog *findReplaceDialog;

// Operations
public:
					// operations (non-undoable)
	void			PerformCut();
	void			PerformPaste();
	void			PerformClear();
	void			PerformTyping(UINT aChar, UINT nRepCnt, UINT nFlags);

					// factory methods for creating actions
	CBEAction	*	MakeEditAction(UINT actionType, void *data = 0);
	CBEAction	*	MakeTypingAction(UINT actionType, void *data = 0);
	CBEAction	*	MakeReplaceAction(UINT actionType, void *data = 0);

	// use whenever the selection changes such that an action, like
	// typing, should finish; normal characters, backspacing, etc.,
	// are _not_ examples, so this is different than EN_SELCHANGE			
	void SelectionChanged();

	// this is thin wrapper around GetSel() which does some adjustments
	// on the endpoints (they didn't always seem so logical, at least
	// for support undo!)
	void GetSelection(int *selStart, int *selEnd) const
	{
		GetEditCtrl().GetSel(*selStart, *selEnd);
		const long len = GetLength();
		if (*selEnd > len || *selEnd == - 1)
			*selEnd = len;
	}

	// a somewhat thin wrapper around SetSel(); here, the problem is
	// the funny 2 character end-of-line sequence; it would be much
	// less of a problem if it were treated consistently, i.e., as
	// a single logical character, but this doesn't always happen
	void SetSelection(int selStart, int selEnd)
	{
		GetEditCtrl().SetSel(selStart, selEnd);
		if (selEnd == selStart + 1)
		{
			CString selText = GetTextSelection();
			if (selText == "\n" && selStart > 0)
				SetSelection(selStart - 1, selEnd);
			else if (selText == "\r" && selEnd < GetLength())
				SetSelection(selStart, selEnd + 1);		
		}
	}

	// provides a uniform interface with CRicherEditView
	int GetLength() const
	{
		return GetBufferLength();
	}

	// ditto here; the base member GetSelectedText() doesn't
	// work well here, as it can return at most a single line
	CString GetTextSelection() const
	{
		int start, end;

		GetSelection(&start, &end);
		LPCTSTR lpszText = ((CEditView*)this)->LockBuffer();
		CString selText(lpszText + start, end - start);

		UnlockBuffer();
		return selText;
	}

	CString CopyTextRange(long start, long end)
	{
		int selStart, selEnd;

		if (start <= end)
		{
			GetSelection(&selStart, &selEnd);

			SetSelection(start, end);
			CString theText = GetTextSelection();

			SetSelection(selStart, selEnd);
			return theText;
		}
		else
			return "";
	}

	void InsertText(LPCTSTR text)
	{
		GetEditCtrl().ReplaceSel(text);
	}

			// support for typing actions
	void	GetItemBefore(int &pos, CTextRange &aRange);
	void	GetItemAfter(int &pos, CTextRange &aRange);

			// turn off redrawing (to avoid flashing
			// when selecting and reselecting text.)
	void	InhibitRedraw(const bool disableRedraw);

			// standard action overrides
	void	OnReplaceSel(LPCTSTR lpszFind, BOOL bNext, BOOL bCase,
						 LPCTSTR lpszReplace);
	void	OnReplaceAll(LPCTSTR lpszFind, LPCTSTR lpszReplace, BOOL bCase);
	BOOL	InitializeReplace(LPCTSTR lpszFind, BOOL bNext, BOOL bCase);

			// by default the action target is this view target; if you
			// change this, you must do it before any actions are created
	void	SetActionTarget(void *at)	{ actionTarget = at; }
	void *	GetActionTarget() const		{ return actionTarget; }

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CBetterEditView)
	protected:
	virtual void OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView);
	//}}AFX_VIRTUAL

// Implementation
protected:

#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

	// Generated message map functions
protected:
	//{{AFX_MSG(CBetterEditView)
	afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnEditClear();
	afx_msg void OnEditCut();
	afx_msg void OnEditPaste();
	afx_msg void OnEditRedo();
	afx_msg void OnUpdateEditRedo(CCmdUI* pCmdUI);
	afx_msg void OnEditUndo();
	afx_msg void OnUpdateEditUndo(CCmdUI* pCmdUI);
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	//}}AFX_MSG
	afx_msg LRESULT OnFindReplaceCmd(WPARAM, LPARAM lParam);
	afx_msg void OnMultipleUndo(UINT nID);
	afx_msg void OnUpdateMultipleUndo(CCmdUI* pCmdUI);
	afx_msg void OnMultipleRedo(UINT nID);
	afx_msg void OnUpdateMultipleRedo(CCmdUI* pCmdUI);
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif
