// RicherEditView.h : header file
//
// Copyright (c) 1998 Matt Clay, all rights reserved.
//

#if !defined(RICHEREDITVIEW_H)
#define RICHEREDITVIEW_H

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

class CREAction;
class CUndoManager;
class CPasteSpecialData;
class COleInsertDialog;

#include "StyleTable.h"
#include "ParagraphTable.h"
#include "TextActionUtil.h"
#include "UndoManager.h"

#include <afxrich.h>
#include <assert.h>

/////////////////////////////////////////////////////////////////////////////
// CRicherEditView view

class CRicherEditView : public CRichEditView
{
public: // create from serialization only
	CRicherEditView();
	DECLARE_DYNCREATE(CRicherEditView)

// Attributes
public:

					// because typing is not atomic, we maintain a
					// last action for this view; most of the time,
					// this will be a typing action, and one which
					// we can continue "typing into"
	CREAction	*	lastAction;

					// last character typed in this edit view
	UINT			lastChar;

					// utility range; subclasses can extend a text range
					// with custom attributes by redefining MakeTextRange();
					// this member var provides a "fast" local; note that
					// it is shared with all callers to GetUtilityTextRange();
					// not used within class CRicherEditView itself, so only
					// your routines; see CRETypingAction::DoBackspace() or
					// DoFwdDelete() for examples of hot to use.)
	CTextRange	*	publicRange;

					// flags for controlling redraws
	bool			redraw;
	DWORD			redrawMask;
	BOOL			locked;

protected:
					// a pointer to the action target; usually a
					// ptr. to the document object of this view
	void			*actionTarget;

					// utility range for this class and its subclasses
	CTextRange	*	privateRange;
	bool			publicRangeAvailable, privateRangeAvailable;

					// are we in insert (default) or overtype mode?
	bool			insertMode;

public:

	static CLIPFORMAT m_cfRTF;

#ifdef ENABLE_DRAG_DROP
	static CLIPFORMAT m_cfPrivate;
	static CLIPFORMAT m_cfEmbeddedObject;
	static CLIPFORMAT m_cfObjectDescriptor;
#endif

	static CFindReplaceDialog *	findReplaceDialog;

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CRicherEditView)
	protected:
	virtual void OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView);
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual			~CRicherEditView();

					// basic edit/clipboard operations
	void			PerformCut();
	void			PerformPaste();
	void			PerformClear();
	void			PerformPasteSpecial(CPasteSpecialData &psd);
	void			PerformTyping(UINT aChar, UINT nRepCnt, UINT nFlags);
	bool			PerformInsertObject(COleInsertDialog &dlg);

					// styling operations
	void			OnStyle(DWORD dwMask, DWORD dwEffects);
	void			OnStyle(DWORD dwMask, DWORD dwEffects, UINT actionType);
	void			PerformStyle(DWORD dwMask, DWORD dwEffects);

	void			PerformColorPick(UINT nID);
	void			PerformColorDefault();

	void			OnParaAlign(WORD wAlign);
	void			OnParaAlign(WORD wAlign, UINT actionType);
	void			PerformParaAlign(WORD align);

	void			OnBullet();
	void			PerformBullet();

	void			OnReplaceSel(LPCTSTR lpszFind, BOOL bNext, BOOL bCase,
								BOOL bWord, LPCTSTR lpszReplace);
	void			OnReplaceAll(LPCTSTR lpszFind, LPCTSTR lpszReplace, BOOL bCase, BOOL bWord);

	void			OnSetCharFormat(const CHARFORMAT &format);
	void			OnSetParaFormat(PARAFORMAT &pf);

					// factory methods for creating actions
	CREAction	*	MakeEditAction(UINT actionType, void *data = 0);
	CREAction	*	MakeTypingAction(UINT actionType, void *data = 0);
	CREAction	*	MakeStyleAction(UINT actionType, void *data = 0);
	CREAction	*	MakeParagraphAction(UINT actionType, void *data = 0);
	CREAction	*	MakeReplaceAction(UINT actionType, void *data = 0);
	CREAction	*	MakeInsertObjectAction(UINT actionType, void *data = 0);

					// factory method for creating text range
	CTextRange	*	MakeTextRange();

					// get the utility range (extremely volatile!)
	CTextRange	&	ReserveUtilityRange()
	{
		assert(publicRange && publicRangeAvailable);
		publicRangeAvailable = false;
		return *publicRange;
	}

	void			ReleaseUtilityRange()
	{
		assert(publicRange && !publicRangeAvailable);
		publicRangeAvailable = true;
		publicRange->resize(0);
	}

#if ENABLE_DRAG_DROP
	CREAction	*	MakeDropAction(UINT actionType, long dropIndex, COleDataObject *pDataObject, BOOL selfDrag);
#endif

	// utility functions; many of these duplicate some aspect
	// of the base class and C(Better)EditView; different names
	// have been chosen 1) for consistency between the edit view
	// subclasses and 2) to avoid overriding non-virtuals, which
	// is always confusing.

	bool InInsertMode() const { return insertMode; }

	// fix up the endpoints which seem to be inconsistent sometimes
	void GetSelection(long *selStart, long *selEnd) const
	{
		GetRichEditCtrl().GetSel(*selStart, *selEnd);
		const long len = GetTextLength();
		if (*selEnd > len || *selEnd == - 1)
			*selEnd = len;
	}

	void SetSelection(long selStart, long selEnd)
	{
		GetRichEditCtrl().SetSel(selStart, selEnd);

		// would like to put end-of-line testing here; both
		// directions are affected but you can't select an
		// individual char in the end-of-line sequence as
		// you can with a plain edit control; hence it is
		// impossible to know which way the selection should
		// be extended.
	}

	long GetLength()
	{
		return GetTextLength();
	}

	CString GetTextSelection() const
	{
		return GetRichEditCtrl().GetSelText();
	}

	void InsertText(LPCTSTR text)
	{
		GetRichEditCtrl().ReplaceSel(text);
	}

	void HideSelection(BOOL hide, BOOL perm)
	{
		GetRichEditCtrl().HideSelection(hide, perm);
	}

public:
			// use whenever the selection changes such that an action, like
			// typing, should finish; normal characters, backspacing, etc.,
			// are _not_ examples, so this is different than EN_SELCHANGE
	void	SelectionChanged();

			// Turn on/off redrawing
	void	InhibitRedraw(const bool disableRedraw);

			// Gets a text element before after the position specified;
			// used to support undoable backspacing and forward delete;
			// note that element can be text, an object, or ...; the 
			// current is preserved
	void	GetElementBefore(long pos, CTextRange &aRange);
	void	GetElementAfter(long pos, CTextRange &aRange);

			// Sets/gets the (character) format; for get, the current
			// selection is used; for set, the endpoints contained in
			// the table are used; can preserve the original selection;
			// generally not extendible as CHARFORMAT, CTextStyle and
			// CStyleTable are not; for extended styles, you need to define
			// additional routines
	void	SetTextStyles(const CTextRange &aRange, bool preserveSelection = true);
	void	GetTextStyles(CTextRange &aRange) const;

protected:
			// gets the style of the element at position i
	void	GetTextStyle(long i, CTextRange &textElement) const;

public:
			// Sets/gets paragraph styles; for get, the current selection
			// is used; for set, the endpoints contained in the table are
			// used;can preserve original selection; generally not extendible
			// as PARAFORMAT, CParagraphStyle, CParagraphTable are not; for
			// extended styles, you need to define additional routines
	void	SetParagraphStyles(const CTextRange &aRange, bool preserveSelection = true);
	void	GetParagraphStyles(CTextRange &aRange) const;

protected:
			// gets the paragraph style of the element at position i
	void	GetParagraphStyle(long i, CTextRange &textElement) const;

public:
			// Inserts/gets text, objects, text styles, and paragraph styles;
			// for get, the current selection is used; for set, the endpoints
			// contained the table are used; the original selection can be
			// preserved
	virtual void InsertTextRange(const CTextRange &aRange, bool preserveSelection = true);
	virtual void GetTextRange(CTextRange &aRange) const;

protected:
			// bottleneck for getting range attributes; override to handle
			// new attributes; as described above, should get information
			// about the text element at position i
	virtual void GetTextRange(long i, CTextRange &textElement) const;

public:
			// inserts (OLE) objects; maintains current selection
	void	InsertObjects(const CTextRange &aRange, bool preserveSelection = true);

protected:
	void	GetObject(long i, CTextRange &textElement) const;

public:
			// by default the action target is this view target; if you
			// change this, you must do it before any actions are created
	void	SetActionTarget(void *at)	{ actionTarget = at; }
	void *	GetActionTarget() const		{ return actionTarget; }

			// Converts a point into a line or character index; character
			// index is for the entire text buffer, not an individual line;
			// also note the character index is strict, unlike the standard
			// tracking routines (if the mouse point is more than midway
			// across a character, the cursor appears after the character)
	long	GetLineIndexFromPoint(const CPoint &aPoint) const;
	long	GetCharIndexFromPoint(const CPoint &aPoint) const;

	void	UpdateUndoMenu();

#if ENABLE_DRAG_DROP
	bool CanUseDataObject(COleDataObject *pDataObject);
	BOOL ImportFromDataObject(COleDataObject *dataObject);
#endif
	
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:
	// get the private utility range (extremely volatile!);
	// only used within this class, and then only withing
	// individual GetXXX() routines
	CTextRange	&ReservePrivateRange()
	{
		assert(privateRangeAvailable);
		privateRangeAvailable = false;
		return *privateRange;
	}

	void ReleasePrivateRange()
	{
		assert(!privateRangeAvailable);
		privateRangeAvailable = true;
		privateRange->resize(0);
	}

protected:
	// Generated message map functions
	//{{AFX_MSG(CRicherEditView)
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnEditClear();
	afx_msg void OnEditCut();
	afx_msg void OnEditPaste();
	afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnCharBold();
	afx_msg void OnCharItalic();
	afx_msg void OnCharUnderline();
	afx_msg void OnParaCenter();
	afx_msg void OnParaLeft();
	afx_msg void OnParaRight();
	afx_msg void OnFormatFont();
	afx_msg void OnOleInsertNew();
	afx_msg void OnEditPasteSpecial();
	afx_msg void OnEditUndo();
	afx_msg void OnUpdateEditUndo(CCmdUI* pCmdUI);
	afx_msg void OnEditRedo();
	afx_msg void OnUpdateEditRedo(CCmdUI* pCmdUI);
	//}}AFX_MSG
	afx_msg void OnColorPick(UINT nID);
	afx_msg void OnColorDefault();
	afx_msg LRESULT OnFindReplaceCmd(WPARAM, LPARAM lParam);
	afx_msg void OnMultipleUndo(UINT nID);
	afx_msg void OnUpdateMultipleUndo(CCmdUI* pCmdUI);
	afx_msg void OnMultipleRedo(UINT nID);
	afx_msg void OnUpdateMultipleRedo(CCmdUI* pCmdUI);
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif
