
// Copyright (c) 1998 Matt Clay, all rights reserved.
//

#if !defined(TEXTACTIONUTIL_H)
#define TEXTACTIONUTIL_H

#include <string>
#include "StyleTable.h"
#include "ParagraphTable.h"
#include "ObjectTable.h"
#include "Action.h"

// This structure can be extended with custom attributes;
// you must override CRicherEditView::MakeTextAction() and
// probably CRicherEditView::GetTextRange(long, CTextRange &)
// as well.

// I prefer using this type of (internal) structure over RTF
// specifically because it's easier

//		* to extend with custom attributes,
//		* to perform common container operations, e.g., append
//		* to get specific range attributes, e.g., just text styles
//		* can be more space-efficient for small and large ranges

// There are drawbacks of course:

//		* Rich Edit Control was designed to use RTF, so selecting
//		  individual characters in sequence seems very slow
//		* you must maintain the "run" structure of the style table.
//		  paragraph table, and any similar tables you define yourself
//		  (assuming you modify them outside of the provided routines)
//		* CTextRange is overkill for the (plain) Edit Control (but it
//		  simplifies the implementation of CBetterEditView and keeps
//		  parity between CBetterEditView and CRicherEditView)

struct CTextRange
{
	CTextRange(const size_t capacity = 0)
		: start(0), end(0), styleTable(capacity),
		  paragraphTable(capacity), objectTable(capacity)
	{ }

	virtual				~CTextRange();

						// resizes all tables as needed creating accessible
						// space for new elements; note that resizing doesn't
						// mean the elements are valid -- you should fill in
						// the space with valid data or resize an individual
						// table accordingly; similiar to the STL's resize();
						// refine when you add tables in a derived class
	virtual void		resize		(size_t n);

						// reserves space in all tables; the capacity of the
						// tables change, not their actual size; similar to the
						// STL's reserve; refine when you add tables in a
						// derived class
	virtual void		reserve		(size_t n);

						// size of the range; because all tables are public,
						// you can access their own size() routine individually
						// if necessary.
	size_t				size()		const		{ return end - start; }
	bool				is_empty()	const		{ return start >= end; }

						// add 'aRange' to the front of this one
	virtual void		prepend		(const CTextRange &aRange);

						// add 'aRange' to the end of this one
	virtual void		append		(const CTextRange &aRange);

	// in dynamic contexts, you should use this member function rather
	// than operator assignment, as the former is not usually virtual
	// and with certain compilers, illegal (a compiler must support
	// covariant return types for that to be workable).
	virtual void		copy		(const CTextRange &aRange);

	// all public; this is a "lightweight" structure
	long				start;
	long				end;
	UMString			text;
	CStyleTable			styleTable;
	CParagraphTable		paragraphTable;
	CObjectTable		objectTable;

protected:
						// individual prepend/append routines; these mainly
						// clarify their parent routines and serve as examples
						// for you to use in a derived class.
	void				prependTextStyles		(const CTextRange &aRange);
	void				prependParagraphStyles	(const CTextRange &aRange);
	void				prependObjects			(const CTextRange &aRange);

	void				appendTextStyles		(const CTextRange &aRange);
	void				appendParagraphStyles	(const CTextRange &aRange);
	void				appendObjects			(const CTextRange &aRange);

};


// utility data structure for style operations
struct StyleActionData
{
	DWORD mask, effect;
};

// an editing key is a destructive key 
// like backspace or forward delete; it
// acts like Edit:Clear
inline bool IsEditingKey(UINT nChar)
{
	return
			(nChar == VK_BACK)
		||
			(nChar == VK_DELETE);
}

// cursor keys are any key which affects
// the location of the cursor, e.g., home
inline bool IsCursorKey(UINT nChar)
{
	return
			(nChar == VK_LEFT)
		||
			(nChar == VK_UP)
		||
			(nChar == VK_RIGHT)
		||
			(nChar == VK_DOWN)
		||
			(nChar == VK_PRIOR)
		||
			(nChar == VK_NEXT)
		||
			(nChar == VK_END)
		||
			(nChar == VK_HOME);
}

// lots of keys...
inline bool IsExtendedKey(UINT nFlags)
{
	return (nFlags & 256) != 0;
}

#endif
