// Copyright (c) 1998 Matt Clay, all rights reserved.
//

#include "stdafx.h"
#include "TextActionUtil.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

CTextRange::~CTextRange() { }

void CTextRange::resize(size_t n)
{
	start = 0;
	end = n;
	text = "";
	styleTable.resize(n);
	paragraphTable.resize(n);
	objectTable.resize(n);
}

void CTextRange::reserve(size_t n)
{
	styleTable.reserve(n);
	paragraphTable.reserve(n);
	objectTable.reserve(n);
}

void CTextRange::copy(const CTextRange &data)
{
	start = data.start;
	end = data.end;

	text = data.text;
	styleTable = data.styleTable;
	paragraphTable = data.paragraphTable;
	objectTable = data.objectTable;
}

void CTextRange::append(const CTextRange &data)
{
	// ignores start, end of data
	// adjust range
	end += data.size();

	// append characters
	text += data.text;

	appendTextStyles(data);
	appendParagraphStyles(data);
	appendObjects(data);
}

// Note the text and paragraph styles are stored in a
// "run" array; within this data structure, common elements
// which are contiquous are not stored individually, but
// rather with a count of how many there are; these routines
// assume that whenever a run array is used, it is correct.

void CTextRange::appendTextStyles(const CTextRange &data)
{
	// append character formats
	CStyleTable::iterator end = styleTable.end();
	CStyleTable::const_iterator first = data.styleTable.begin();

	if (styleTable.begin() != end)
	{
		CStyleTable::reference last = styleTable.back();

		// If the beginning of the source range matches the end of the
		// the destination, merge them. Otherwise, the remaining runs
		// must be different so we can just copy them.

		if (first && *first == last)
		{
			last.count += (*first).count;
			++first;
		}	
	}

	styleTable.insert(end, first, data.styleTable.end());
}

void CTextRange::appendParagraphStyles(const CTextRange &data)
{
	// append paragraph formats
	CParagraphTable::iterator end = paragraphTable.end();
	CParagraphTable::const_iterator first = data.paragraphTable.begin();

	if (paragraphTable.begin() != end)
	{
		CParagraphTable::reference last = paragraphTable.back();

		// If the beginning of the source range matches the end of the
		// the destination, merge them. Otherwise, the remaining runs
		// must be different so we can just copy them.

		if (first && *first == last)
		{
			last.count += (*first).count;
			++first;
		}	
	}

	paragraphTable.insert(end, first, data.paragraphTable.end());
}

void CTextRange::appendObjects(const CTextRange &data)
{
	// append the objects as they must be unique
	objectTable.insert(objectTable.end(), data.objectTable.begin(), data.objectTable.end());
}


void CTextRange::prepend(const CTextRange &data)
{
	// adjust range
	start -= data.size();

	// prepend characters
	text = data.text + text;
}

void CTextRange::prependTextStyles(const CTextRange &data)
{
	// prepend character formats
	CStyleTable::iterator first = styleTable.begin();
	CStyleTable::const_iterator end = data.styleTable.end();

	if (first != styleTable.end() && data.styleTable.begin() != end)
	{
		CStyleTable::const_reference last = data.styleTable.back();

		// If the end of the source range matches the beginning of the
		// the destination, merge them. Otherwise, the remaining runs
		// must be different so we can just copy them.

		if (first && *first == last)
		{
			(*first).count += last.count;
			--end;
		}	
	}

	styleTable.insert(first, data.styleTable.begin(), end);
}

void CTextRange::prependParagraphStyles(const CTextRange &data)
{
	// prepend character formats
	CParagraphTable::iterator first = paragraphTable.begin();
	CParagraphTable::const_iterator end = data.paragraphTable.end();

	if (first != paragraphTable.end() && data.paragraphTable.begin() != end)
	{
		CParagraphTable::const_reference last = data.paragraphTable.back();

		// If the end of the source range matches the beginning of the
		// the destination, merge them. Otherwise, the remaining runs
		// must be different so we can just copy them.

		if (first && *first == last)
		{
			(*first).count += last.count;
			--end;
		}	
	}

	paragraphTable.insert(first, data.paragraphTable.begin(), end);
}

void CTextRange::prependObjects(const CTextRange &data)
{
	// just append the objects as they must be unique
	objectTable.insert(objectTable.begin(),
						data.objectTable.begin(),
						data.objectTable.end());
}
